#!/usr/bin/env node
'use strict';

// Dependencies
const _ = require('lodash');
const $ = require('highland');
const util = require('util');
const yargs = require('yargs');
const files = require('../lib/files');
const functions = require('../lib/functions');
const errors = require('../lib/errors');
const logging = require('../lib/logging');
const Action = require('../lib/actions');
const program = require('../lib/program');
const Promise = require('bluebird');
const microtime = require('microtime');

// Logger
const logger = logging.createLogger('nodus-run');

// Program Help
yargs
    .usage('Usage: $0 <command> [args] [options]')
    .version()
    .epilog('Copyright (c) 2017 Brad Serbu <bradserbu@bradserbu.com>\nAll rights reserved.')
    .help();

function isStream(value) {
    return util.isFunction(value.pipe);
}

function print_result(value) {

    if (util.isNullOrUndefined(value))
        return; // Print Nothing

    if (isStream(value)) {
        return new Promise((resolve, reject) =>
            $(value)
                .each(result => print_result(result))
                .stopOnError(err => reject(err))
                .done(() => resolve())
        );
    }

    // Write a string as is to the console
    if (util.isString(value)) {
        return process.stdout.write(value + '\n');
    }

    // Write the results as a JSON string
    if (util.isObject(value)) {
        // TODO: Add indent option to CLI args
        return print_result(JSON.stringify(value, null, 2));
    }

    // Write the string representation of the value
    process.stdout.write(value.toString() + '\n');
}

function print_error(err) {
    console.error(err);
}

/**
 * Load a command to run from the target
 * @param target
 * @returns {*}
 */
function loadActivity(targetFile) {
    // Load the module/file to be run
    const target = files.requireFile(targetFile);
    logger.debug('[TARGET]', target);

    const name = target.name || targetFile;
    logger.debug('[TARGET_NAME]', name);

    return Action(name, target);
}

// Parse CommandLine Arguments
const argv = yargs.argv;
logger.debug('[ARGV]', argv);

// Determine the file/module to be run
const targetFile = argv._.shift();
logger.debug('[TARGET_FILE]', targetFile);

// Display help if no target is specified
if (util.isNullOrUndefined(targetFile)) {
    return yargs.showHelp();
}

// Load command to run
const activity = loadActivity(targetFile);


// Log Events
const started_on = microtime.now();
activity.onAny((event, data) => {

    const elapsed_ms = ((microtime.now() - started_on) * .001).toFixed(2);

    logger.debug(`EVENT:${event}`, data, `(${elapsed_ms} ms)`);
    // logger.debug(`[${event.toUpperCase()}]`, data, `(${elapsed_ms} ms)`)
});

// Run the activity and display the result
activity
    .mappedNamedArgs()
    .run(argv)  // TODO: Support ability to interpet args from options
    .catch(err => print_error(err))
    .then(result => print_result(result))
    .then(() => program.shutdown());